/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Annotation.hxx
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.05.2014
 *  Declaraion of parameter class Annotation.
 */

#ifndef _ANNOTATION_HXX_
#define _ANNOTATION_HXX_

#include <algorithm>

//#include <CImg.h>
#include <stdexcept>
#include <iostream>
#include <iomanip>
#include <list>

namespace annotation {

	/// Annotation class which keeps all common parameters for sub annotation classes
	class Annotation;

	/// Annotation box element keeps two coordinates for upper left and lower right points
	class AnnotationBox;

	/// Frame objects keep id, filename and pointers to all corresponding annotation objects
	class Frame;


	/// Keeps all Sequence objects
	class AnnotationList {

	public:		
		/// add frame if it does not exist already
		void addFrame( Frame _frame );

		/// add annotation
		int addAnnotationBox( AnnotationBox _annotation );

		/// return pointer to according frame
		const Frame* getFrame( std::string _filename ) const;

		/// return pointer to according annotation
		const AnnotationBox* getAnnotationBox( int _annotationId ) const;

		/// get annotations list of according Frame
		const std::vector< const AnnotationBox* > getAnnotationBoxes( std::string _filename );

		/// return vector of all frames
		const std::vector< const Frame* > getFrames() const;

		/// delete annotation and annotationId in corresponding frame object
		bool deleteAnnotation( int _annotationId );

	private:
		/// A map which maps filenames to frame objects
		std::map< std::string, Frame > frames;

		/// A map which maps annatation ids to annotation box objects
		std::map< int, AnnotationBox > annotationBoxes;

	};

	/// Annotation class which keeps all common parameters for sub annotation classes
	class Annotation
	{
	public:

		/// Id of annotation object, which is unique for a specific frame
		int annotationId ;

		/// Frame filename
		std::string filename;

		/// flag for internally deleted annotations
		bool deleted;

		/// flag for finalized annotations
		bool finalized;

		/// state of task result
		std::string taskResultState;

		/// task result id
		std::string taskResultId;

		/// default constructor
		Annotation(std::string _filename, int _annotationId = -1, bool _deleted = false, bool _finalized = false, std::string _taskResultId = "", std::string _taskResultState = "") :
			filename(_filename), annotationId(_annotationId), deleted(_deleted), finalized(_finalized), taskResultId(_taskResultId), taskResultState(_taskResultState)
		{
		}

		virtual ~Annotation() {} ;

	} ;


	/// class for describing an annotation element (box, contour...)
	class AnnotationBox : public Annotation
	{
	public:

		/// Id of annotation object, which is unique for a specific frame
		float x1, x2, y1, y2;

		/// default constructor
		AnnotationBox(std::string _filename, float _x1,  float _y1,  float _x2,  float _y2 ) : 
			Annotation( _filename ), x1(_x1), x2(_x2), y1(_y1), y2(_y2)
		{}

		// constructor if annotation id is known
		AnnotationBox(std::string _filename, int _annotationId, float _x1, float _y1, float _x2, float _y2) :
			Annotation(_filename, _annotationId), x1(_x1), x2(_x2), y1(_y1), y2(_y2)
		{}

		// constructor if annotation id and state (deleted, final) is known
		AnnotationBox(std::string _filename, bool _deleted, bool _finalized, float _x1, float _y1, float _x2, float _y2) :
			Annotation(_filename, -1, _deleted, _finalized), x1(_x1), x2(_x2), y1(_y1), y2(_y2)
		{}

		// constructor if annotation id and state (deleted, final) is known plus task result id
		AnnotationBox(std::string _filename, bool _deleted, bool _finalized, std::string _taskResultId, float _x1, float _y1, float _x2, float _y2) :
			Annotation(_filename, -1, _deleted, _finalized, _taskResultId), x1(_x1), x2(_x2), y1(_y1), y2(_y2)
		{}

		// constructor if state (deleted, final)
		AnnotationBox(std::string _filename, int _annotationId, bool _deleted, bool _finalized, float _x1, float _y1, float _x2, float _y2) :
			Annotation(_filename, _annotationId, _deleted, _finalized), x1(_x1), x2(_x2), y1(_y1), y2(_y2)
		{}

		// constructor with all parameters known
		AnnotationBox(std::string _filename, int _annotationId, bool _deleted, bool _finalized, std::string _taskResultId, float _x1, float _y1, float _x2, float _y2) :
			Annotation(_filename, _annotationId, _deleted, _finalized, _taskResultId), x1(_x1), x2(_x2), y1(_y1), y2(_y2)
		{}

		// constructor with all parameters known
		AnnotationBox(std::string _filename, int _annotationId, bool _deleted, bool _finalized, std::string _taskResultId, std::string _taskResultState, float _x1, float _y1, float _x2, float _y2) :
			Annotation(_filename, _annotationId, _deleted, _finalized, _taskResultId, _taskResultState), x1(_x1), x2(_x2), y1(_y1), y2(_y2)
		{}

		virtual ~AnnotationBox() {} ;

	} ;

	
	/// Frame objects keep id, filename and pointers to all corresponding annotation objects
	class Frame{
	public:
		/// Frame id
		int frameId;

		/// Frame filename
		std::string filename;

		/// pointers to all corresponding annotation elements
		std::vector< int > annotationIds;

		/// return pointer to frame with frameId
		//Annotation* operator()(size_t annotationId);
		
		Frame ( std::string _filename, int _frameId = -1 ) : 
			frameId(_frameId), filename(_filename){}

		//Frame ( int _frameId, std::string _filename, std::vector< int > &_annotationIds ) : 
		//	frameId(_frameId), filename(_filename), annotationIds(_annotationIds){}


	};


	/// add frame if it does not exist already
	void AnnotationList::addFrame( Frame _frame ){
		frames.insert( std::pair< std::string, Frame >( _frame.filename, _frame ) );
	}


	/// add annotation
	int AnnotationList::addAnnotationBox( AnnotationBox _annotation ){

		// check if corresponding frame exists, if not, throw error since we could
		// add this frame but do not know its parameters beside the frame id
		std::map<std::string,Frame>::iterator it = frames.find( _annotation.filename );
				
		int newId = 0;

		if (it == frames.end()){
			Frame frame(_annotation.filename);
			addFrame(frame);
			it = frames.find(_annotation.filename);
		}

		if (it != frames.end()){
			
			// if _annotation.Id == -1, apply new id 
			if( _annotation.annotationId == -1 ){
				if( !annotationBoxes.empty() )
					newId = annotationBoxes.rbegin()->first + 1;
				_annotation.annotationId = newId;
			}
			else
				newId = _annotation.annotationId;

			std::pair< std::map<int,AnnotationBox>::iterator, bool > ret;
			ret = annotationBoxes.insert ( std::pair< int, AnnotationBox >( newId, _annotation ) );
			
			// if annotation does not exist already, add id to according frame
			if (ret.second!=false) {
				it->second.annotationIds.push_back( _annotation.annotationId );
			}
			else
				throw( std::range_error("AnnotationList::addAnnotationBox: annotationId already exists") );

		}
		// frame does not exist, throw error
		else{
			throw( std::range_error("AnnotationList::addAnnotationBox: Frame does not exist") );
		}

		return newId;

	}


	const Frame* AnnotationList::getFrame( std::string _filename ) const {
		const std::map<std::string,Frame>::const_iterator it = frames.find( _filename );
		if( it != frames.end() ){
			return &(it->second);
		}
		else
			throw( std::range_error("AnnotationList::getFrame: frameId out of range") );
	}

	
	/// return vector of all frames
	const std::vector< const Frame* > AnnotationList::getFrames() const {
	
		std::vector< const Frame* > vec;
	
		auto it = frames.begin();
		for( ; it != frames.end(); it++ ){
			vec.push_back(  &(it->second) );
		}
	
		return vec;
	
	}

	const AnnotationBox* AnnotationList::getAnnotationBox( int annotationId ) const {
		std::map<int,AnnotationBox>::const_iterator it = annotationBoxes.find( annotationId );
		if( it != annotationBoxes.end() )
			return &(it->second);
		else
			throw( std::range_error("AnnotationList::getAnnotationBox: annotationId out of range") );
	}


	const std::vector<const AnnotationBox*> AnnotationList::getAnnotationBoxes( std::string _filename ){
	
		// get frame
		std::map<std::string,Frame>::iterator it = frames.find( _filename );
		if( it == frames.end() )
			throw( std::range_error("AnnotationList::getAnnotationBoxes: frameId out of range") );

		std::vector< const AnnotationBox* > vec;
	
		for( auto it2 = it->second.annotationIds.begin(); it2 != it->second.annotationIds.end(); it2++ ){
			vec.push_back( getAnnotationBox( *it2 ) );
		}
		return vec; //it->second.annotationIds;
	
	}

} ; /*namespace annotation */


#endif /* #ifndef _ANNOTATION_HXX_ */
