/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file BlurByDepth.hxx
 *  Implementation of the parameter class BlurByDepth.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 07.05.2014
 */

#ifndef _BLURBYDEPTH_HXX_
#define _BLURBYDEPTH_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "BlurByDepth.h"

template <typename T>
BlurByDepth<T>::BlurByDepth(const std::string& name) :
		TemplatedParameteredObject<T>(
			"BlurByDepth", name,
			"<h2>Blurs an image depending on depths of a given depth "
			"map</h2><br>"
			"Blurs an image depending on depths of a given depth map"
		)
{
	ParameteredObject::_addInputSlot(
		images, "images",
		"Images to blur",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		depthmaps, "depthmaps",
		"Depthmaps for depth dependen blur",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		blurred, "blurred",
		"Blurred images",
		"CImgList<T>");

	ParameteredObject::_addParameter< bool >(
		blurQuadratic, "blurQuadratic",
		"Quadratic scheme. Near pixels are blurred stronger by quadratic weight.",
		false, "bool");
	ParameteredObject::_addParameter< float >(
		depthNear, "depthNear",
		"depth to start blur with blurNear",
		1.0, "float");
	ParameteredObject::_addParameter< float >(
		depthFar, "depthFar",
		"depth to end blur with blurFar",
		0.0, "float");
	ParameteredObject::_addParameter< float >(
		blurNear, "blurNear",
		"blur kernel for nearest points",
		20, "float");
	ParameteredObject::_addParameter< float >(
		blurFar, "blurFar",
		"blur kernel for farest point",
		5, "float");
	ParameteredObject::_addParameter< bool >(
		stereo, "stereo",
		"Assume two images per depth map",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		useMaxMinValues, "useMaxMinValues",
		"Use given max and min values for all frames",
		false, "bool");
	ParameteredObject::_addParameter< float >(
		maxValue, "maxValue",
		"Maxmimum value of all images",
		255, "float");
	ParameteredObject::_addParameter< float >(
		minValue, "minValue",
		"Minimum value of all images",
		0, "float");

}

template <typename T>
void BlurByDepth<T>::execute() {
	
	const cimg_library::CImgList<T> &img = this->images();
	const cimg_library::CImgList<T> &dm = this->depthmaps();
	float blurNear = this->blurNear() ;
	float blurFar = this->blurFar() ;
	float blurRange = blurFar - blurNear ;

	// number of depths in front and behind current pixel which are considered to use
	// for blur. All pixels of current image patch which ly inside of this depth range are 
	// used, to calculate the blur value for the pixel in patch center.
	// Value is relative to depth range.
	float blurDepthRange = 0.1;
	// min percentage of patch pixels used to calculate blur. If the number of chosen pixels
	// is smaller than this value, double 'blurDepthRange' and try again.
	float nImpact = 0.5;
	
	if( img.size() != dm.size() && !this->stereo() )
		throw( std::range_error("BlurByDepth<T>::execute: depthmaps.size != images.size") );
	if( img.size() != 2*dm.size() && this->stereo() )
		throw( std::range_error("BlurByDepth<T>::execute: STEREO-MODE: depthmaps.size != 2* images.size") );

	float maxVal = 0;

	for( int i = 0; i < img.size(); i++ ){

		sout << "Image " << i << ":" << std::endl;

		// In this code, taller values of depth maps mean nearer pixels!

		float depthMax = dm[i].max() ;	
		float depthMin = dm[i].min() ;

		if( this->useMaxMinValues() ){

			maxVal = std::max( maxVal, depthMax );
			
			sout << "Replace max value of image (" << depthMax << ") by " << this->maxValue() << std::endl;
			sout << "Replace min value of image (" << depthMin << ") by " << this->minValue() << std::endl;

			depthMax = this->maxValue();
			depthMin = this->minValue();

		}

		depthMax *= this->depthNear() ;	
		depthMin += this->depthFar() * depthMax;
		float depthRange = depthMax - depthMin ;

		sout << "Near value: " << depthMax << " / far value: " << depthMin << std::endl;
				
		cimg_library::CImg<T> blur = img[i];
		blur.fill(255);
		const int w = blur.width() ;
		const int h = blur.height() ;
		const int spectrum = blur.spectrum() ;
				
		for( int c = 0; c < spectrum; c++ ){
			for( int y = 0; y < h; y++ ){

				if( y%(h/3) == 0 )
					sout << "\tprogress: " << int(float(y)/float(h)*100.) << "%" << std::endl;

				#pragma omp parallel for
				for( int x = 0; x < w; x++ ){
				
					// get kernel
					float pDepth = std::max( depthMin, std::min( depthMax, (float)dm[i](x,y,0,0) ) );	// depth of pixel to blur
					float sigma = blurNear;
					if( dm[i](x,y,0,0) != 0 ){
						float c1 = 1 - ( pDepth - depthMin ) / depthRange ;	// choose sigma for blur
						if( this->blurQuadratic() ) 
							c1 *= c1;
						sigma = blurNear + c1 * blurRange ;
					}
				
					// get patch
					int x0 = std::max( 0, x - int( sigma + 0.5 ) );
					int x1 = std::min( w-1, x + int( sigma + 0.5 ) );
					int y0 = std::max( 0, y - int( sigma + 0.5 ) );
					int y1 = std::min( h-1, y + int( sigma + 0.5 ) );
					int pSize = (x1-x0+1)*(y1-y0+1);

					float pBlur = 0;	// blurred value
					int pNum = 0;		// number of chosen pixel for blur
					float pBlurDepthRange = blurDepthRange * depthRange;

					float pMinDepth = pDepth - pBlurDepthRange;
					float pMaxDepth = pDepth + pBlurDepthRange;

					//sout << "pSize: " << pSize << std::endl;

					//sout << "get patch blur" << x << "," << y << std::endl;

					// Do until enough neighbour pixel are chosen for blur
					int iter = 1;
					while( float(pNum) < float(nImpact * pSize) ){
						pNum = 0;
						pBlur = 0;
						// iterate over patch
						for( int px = x0; px <= x1; px++ ){
							for( int py = y0; py <= y1; py++ ){
								// check if current pixel is inside of given depth range
								float pDepthN = std::max( depthMin, std::min( depthMax, (float)dm[i](px,py,0,0) ) );	// depth of current pixel
								if( pDepthN >= pMinDepth && pDepthN <= pMaxDepth ){
									pNum ++;
									pBlur += img[i](px,py,0,c);
								}
							}
						}

						//if(iter > 1 ){
						//	sout << float(pNum) << std::endl;
						//	sout << float(nImpact * pSize) << std::endl;
						//
						//	sout << "Try(" << iter << "):" << std::endl;
						//	sout << "pMinDepth: " << pMinDepth << std::endl;
						//	sout << "pMaxDepth: " << pMaxDepth << std::endl;
						//}

						iter++;

						pMinDepth -= pBlurDepthRange;
						pMaxDepth += pBlurDepthRange;
					}

					blur(x,y,0,c) = pBlur / float(pNum);	// save blurred pixel
					
				}
			}
		}

		sout << std::endl;

		this->blurred().push_back( blur );

	}

	sout << "--------------------" << std::endl;
	sout << "Maximal value of all images: " << maxVal << std::endl << std::endl;

}

#endif /* _BLURBYDEPTH_HXX_ */
