/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DrawBoxes.hxx
 *  Implementation of the parameter class DrawBoxes.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.05.2014
 */

#ifndef _DRAWBOXES_HXX_
#define _DRAWBOXES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "DrawBoxes.h"

template <typename T>
DrawBoxes<T>::DrawBoxes(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DrawBoxes", name,
			"<h2>Draw boxes on given images</h2><br>"
			"Draw boxes on given images"
		)
{
	ParameteredObject::_addInputSlot(
		annotations, "annotations",
		"Annotation list object",
		"AnnotationList");
	ParameteredObject::_addInputSlot(
		filepaths, "filepaths",
		"filepaths",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		images, "images",
		"images",
		"CImgList<T>");

	ParameteredObject::_addParameter(
		processAll, "processAll",
		"Save all input images whether with or without drawn boxes."
		"Otherwise, only images with boxes are saved.", false);

	ParameteredObject::_addParameter(
		resultsFinal, "resultsFinal",
		"draw final results", true);

	ParameteredObject::_addParameter(
		resultsAdd, "resultsAdd",
		"draw add results", true);

	ParameteredObject::_addParameter(
		resultsDone, "resultsDone",
		"draw done results", true);

	ParameteredObject::_addParameter(
		resultsFailed, "resultsFailed",
		"draw failed results", true);

	ParameteredObject::_addParameter(
		resultsEdit, "resultsEdit",
		"draw edit results", true);

	ParameteredObject::_addParameter(
		colorR, "colorR", "Color: red (0-255)", 255u);
	ParameteredObject::_addParameter(
		colorG, "colorG", "Color: green (0-255)", 255u);
	ParameteredObject::_addParameter(
		colorB, "colorB", "Color: blue (0-255)", 0u);

	ParameteredObject::_addParameter(
		colorRWrong, "colorRWrong", "Wrong Color: red (0-255)", 255u);
	ParameteredObject::_addParameter(
		colorGWrong, "colorGWrong", "Wrong Color: green (0-255)", 0u);
	ParameteredObject::_addParameter(
		colorBWrong, "colorBWrong", "Wrong Color: blue (0-255)", 0u);

	ParameteredObject::_addParameter(
		colorRFinal, "colorRFinal", "Final Color: red (0-255)", 0u);
	ParameteredObject::_addParameter(
		colorGFinal, "colorGFinal", "Final Color: green (0-255)", 255u);
	ParameteredObject::_addParameter(
		colorBFinal, "colorBFinal", "Final Color: blue (0-255)", 0u);

	ParameteredObject::_addParameter(
		colorRFinalWrong, "colorREdit", "Final Wrong Color: red (0-255)", 0u);
	ParameteredObject::_addParameter(
		colorGFinalWrong, "colorGEdit", "Final Wrong Color: green (0-255)", 0u);
	ParameteredObject::_addParameter(
		colorBFinalWrong, "colorBEdit", "Final Wrong Color: blue (0-255)", 255u);
}

std::string getFilepath( std::string filename, const std::vector<std::string> &list ){

	std::string path = "";
	std::string filename2 = "";
	for( auto it = list.begin(); it != list.end(); it++ ){
		filename2 = (*it).substr((*it).find_last_of(FileTool::slash)+1,(*it).length());
		if( filename2 == filename )
			path = (*it);
	}

	//sout << "filename2 = " << filename2 << std::endl;

	return path;

}

template <typename T>
void DrawBoxes<T>::execute() {
	// check uint8 range
	assert (colorR() < 256u);
	assert (colorG() < 256u);
	assert (colorB() < 256u);

	// paint color
	unsigned char color[] = {
		(unsigned char) this->colorR(),
		(unsigned char) this->colorG(),
		(unsigned char) this->colorB()
	};

	// paint wrong color
	unsigned char color2[] = {
		(unsigned char) this->colorRWrong(),
		(unsigned char) this->colorGWrong(),
		(unsigned char) this->colorBWrong()
	};

	// paint final color
	unsigned char color3[] = {
		(unsigned char) this->colorRFinal(),
		(unsigned char) this->colorGFinal(),
		(unsigned char) this->colorBFinal()
	};

	// paint final color
	unsigned char color4[] = {
		(unsigned char) this->colorRFinalWrong(),
		(unsigned char) this->colorGFinalWrong(),
		(unsigned char) this->colorBFinalWrong()
	};

	const annotation::AnnotationList &in = this->annotations();

	const std::vector< const annotation::Frame* > frames = in.getFrames();

	int nBoxesAll = 0;

	// iterate over list of all available frames
	auto it = frames.begin();
	for (; it != frames.end(); it++ ){

		// get filename saved in annoation list
		std::string filename = (*it)->filename;

		// find corresponding filepath from filelist
		std::string filepath = getFilepath(filename, this->filepaths());

		if( filepath == "" ){
			sout << "Could not find image: " << filename << std::endl;
			continue;
		}
		
		cimg_library::CImg<T> img( filepath.c_str() );
		img.resize(img.width(),img.height(),img.depth(),3);
		//cimg_library::CImg<T> img(2000,1000,1,3,0);

		const int w = img.width();
		const int h = img.height();

		int nBoxes = 0;

		
		
		// iterate over list of annotations
		auto it2 = (*it)->annotationIds.begin();
		const auto it2End = (*it)->annotationIds.end();
		for (; it2 != it2End; it2++){

			const annotation::AnnotationBox* box = in.getAnnotationBox(*it2);
			int x1 = std::max(0, std::min(w - 1, (int)box->x1));
			int y1 = std::max(0, std::min(h - 1, (int)box->y1));
			int x2 = std::max(0, std::min(w - 1, (int)box->x2));
			int y2 = std::max(0, std::min(h - 1, (int)box->y2));
			
			nBoxes++;

			unsigned char *col = &color[0];

			if (box->taskResultState == "Done" && this->resultsDone())
				col = &color3[0];
			else if (box->taskResultState == "Edit" && this->resultsEdit())
				col = &color4[0];
			else if (box->taskResultState == "Add" && this->resultsAdd())
				col = &color[0];
			else if (box->taskResultState == "Failed" && this->resultsFailed())
				col = &color2[0];
			else if (box->taskResultState == "Final" && this->resultsFinal()){
				col = &color3[0];
				img.draw_line(x1, y1, x2, y2, col);
				img.draw_line(x2, y1, x1, y2, col);
			}
			else{
				continue;
			}

			img.draw_line(x1, y1, x2, y1, col);	// top
			img.draw_line(x2, y1, x2, y2, col);	// right
			img.draw_line(x1, y2, x2, y2, col);	// bottom
			img.draw_line(x1, y1, x1, y2, col);	// left				


		}
		
		// Only save images if at least one box was painted to it
		if (nBoxes > 0 || this->processAll() ){
			sout << "\tProcessed '" << filepath << "'" << std::endl;
			this->images().push_back( img );
		}

		nBoxesAll += nBoxes;			

	}

	sout << "Painted " << nBoxesAll << " boxes. Thank you and see you again soon!" << std::endl;
	
	
	//const annotation::AnnotationList &list = this->annotations();
	
	//const annotation::Frame* frame = list.getFrame("wer");
	//
	//sout << frame->frameId << std::endl;
	//
	//sout << list.getFrame(0)->frameId << std::endl;
	
	//for( auto it = list.begin(); it != list.end(); it++ ){
	//	sout << it->frameId << ": " << it->annotations.size() << std::endl;
	//}
	//
	//for( auto it = box.begin(); it != box.end(); it++ ){
	//	sout << it->annotationId << ": " << it->frame->filename << std::endl;
	//}

}

#endif /* _DRAWBOXES_HXX_ */
