/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file AnnotationReader.cpp
 *  Implementation of parameter class AnnotationReader.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 24.04.2014
 */

#include <charon-core/ParameteredObject.hxx>
#include <charon-utils/AnnotationReader.h>
#include <fstream>

AnnotationReader::AnnotationReader(const std::string& name) :
		ParameteredObject(
			"AnnotationReader", name,
			"<h2>Reads JSON box values</h2><br>Reads JSON box values and saves coordinates into std vectors."
		)
{

	ParameteredObject::_addOutputSlot(
		annotations, "annotations",
		"List of frame objects containing frameids, filenames etc.",
		"AnnotationList");

	ParameteredObject::_addParameter(
		jsonFile, "jsonFile",
		"JSON file",
		"fileopen");

	ParameteredObject::_addParameter<int>(
		maxObjects, "maxObjects",
		"Maximal objects to read in",
		0,
		"int");

	ParameteredObject::_addParameter<bool>(
		verbose, "verbose",
		"Extended log",
		false,
		"bool");

}

void AnnotationReader::execute() {
	
	annotation::AnnotationList & out = this->annotations();
		
	// Read in file and convert it to json object
	std::ifstream file;
	file.open ( this->jsonFile() );
	std::string s((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());
	file.close();

	json::Value json = json::Deserialize( s );

	//sout << json.GetType() << std::endl;

	// get root object
	json::Array arr = json.ToArray();

	size_t nJSON = arr.size();

	sout << nJSON << std::endl;

	if( this->maxObjects() > 0 )
		nJSON = std::min( (int) arr.size(), this->maxObjects() );

	sout << "Read " << nJSON << " annotations" << std::endl;

	if (this->verbose()) {
		sout << "frameId \tannotationId \tisDeleted \tisFinal \tx1 \ty1 \tx2 \ty2 \tfilename " << std::endl;
		sout << "-- \t-- \t-- \t-- \t-- \t-- \t-- \t-- \t-- " << std::endl;
	}

	// iterate through first array
	for( size_t i = 0; i < nJSON; i++ ){

		json::Object frameData = arr[i].ToObject();
		
		int frameId = frameData.find("frame")->second.ToInt();
		std::string filename = frameData.find("filename")->second.ToString();
		
		// Read in new frame
		char *buffer = new char[filename.length()+1];
		sprintf (buffer,filename.c_str(),frameId);
		filename = std::string(buffer);

		// add frame to annotation list
		out.addFrame( annotation::Frame( filename, frameId ) );

		json::Array frameBoxes = frameData.find("results")->second.ToArray();
		
		//// save points of a box to vector
		for( size_t j = 0; j < frameBoxes.size(); j++ ){

			json::Object boxData = frameBoxes[j].ToObject();
			
			int aid = boxData.find("id")->second.ToInt();	// annotation id from database
			
			bool isDeleted = boxData.find("isDeleted")->second.ToBool();	// deleted state
			
			bool isFinal = boxData.find("isFinal")->second.ToBool();	// final state
			
			std::string taskResultState = boxData.find("taskResultState")->second.ToString();	// final state
			
			std::string taskResultId = boxData.find("taskResultId")->second.ToString();	// final state

			if (isDeleted && isFinal && this->verbose() )
				sout << "\tDELETED & FINAL!" << std::endl;
			
			json::Array boxCoords = boxData.find("data")->second.ToArray();
			
			float y1 = boxCoords[0].ToFloat() ;
			float x1 = boxCoords[1].ToFloat() ;
			float y2 = boxCoords[2].ToFloat() ;
			float x2 = boxCoords[3].ToFloat() ;

			aid = out.addAnnotationBox(annotation::AnnotationBox(filename, aid, isDeleted, isFinal, taskResultId, taskResultState, x1, y1, x2, y2));
			
			if( this->verbose() ) {
				sout << frameId << "\t" << aid << "\t" << isDeleted << "\t" << isFinal << "\t" << x1 << "\t" << y1 << "\t" << x2 << "\t" << y2 << "\t" << filename << std::endl;
			}
			
		}

	}

}

// the following functions are needed
// for class AnnotationReader to work as a charon plugin.
extern "C" annotationreader_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new AnnotationReader(name);
}

extern "C" annotationreader_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" annotationreader_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
